/* One way encryption based on SHA512 sum.
   Copyright (C) 2007, 2009 Free Software Foundation, Inc.
   This file is part of the GNU C Library.
   Contributed by Ulrich Drepper <drepper@redhat.com>, 2007.
   Optimized for speed by Milen Rangelov <gat3way@gat3way.eu>, 2010

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */


#define _GNU_SOURCE
#define _XOPEN_SOURCE 600
#include <assert.h>
#include <errno.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/param.h>
#include "plugin.h"



static const char sha512_salt_prefix[] = "$6$";

static const char sha512_rounds_prefix[] = "rounds=";

/* Maximum salt string length.  */
#define SALT_LEN_MAX 16
/* Default number of rounds if not explicitly specified.  */
#define ROUNDS_DEFAULT 5000
/* Minimum number of rounds.  */
#define ROUNDS_MIN 1000
/* Maximum number of rounds.  */
#define ROUNDS_MAX 999999999

/* Table with characters for base64 transformation.  */
static const char b64t[64] =
"./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";


/* Prototypes for local functions.  */
extern char *__sha512_crypt_r (const char *key, const char *salt,
			       char *buffer, int buflen);
extern char *__sha512_crypt (const char *key, const char *salt);


char *
__sha512_crypt_r (key, salt, buffer, buflen)
     const char *key;
     const char *salt;
     char *buffer;
     int buflen;
{
  unsigned char alt_result[64]
    __attribute__ ((__aligned__ (__alignof__ (uint64_t))));
  unsigned char temp_result[64]
    __attribute__ ((__aligned__ (__alignof__ (uint64_t))));
  size_t salt_len;
  size_t key_len;
  size_t cnt;
  char *cp;
  char *copied_key = NULL;
  char *copied_salt = NULL;
  char *p_bytes;
  char *s_bytes;
  /* Default number of rounds.  */
  size_t rounds = ROUNDS_DEFAULT;
  bool rounds_custom = false;
  char *bigbuf, *bigbuf2;
  int bbl, bbl2;

  /* Find beginning of salt string.  The prefix should normally always
     be present.  Just in case it is not.  */
  if (strncmp (sha512_salt_prefix, salt, sizeof (sha512_salt_prefix) - 1) == 0)
    /* Skip salt prefix.  */
    salt += sizeof (sha512_salt_prefix) - 1;

  if (strncmp (salt, sha512_rounds_prefix, sizeof (sha512_rounds_prefix) - 1)
      == 0)
    {
      const char *num = salt + sizeof (sha512_rounds_prefix) - 1;
      char *endp;
      unsigned long int srounds = strtoul (num, &endp, 10);
      if (*endp == '$')
        {
          salt = endp + 1;
          rounds = MAX (ROUNDS_MIN, MIN (srounds, ROUNDS_MAX));
          rounds_custom = true;
        }
    }

  salt_len = MIN (strcspn (salt, "$"), SALT_LEN_MAX);
  key_len = strlen (key);

  if ((key - (char *) 0) % __alignof__ (uint64_t) != 0)
    {
      char *tmp = (char *) alloca (key_len + __alignof__ (uint64_t));
      key = copied_key =
        memcpy (tmp + __alignof__ (uint64_t)
                - (tmp - (char *) 0) % __alignof__ (uint64_t),
                key, key_len);
      assert ((key - (char *) 0) % __alignof__ (uint64_t) == 0);
    }

  if ((salt - (char *) 0) % __alignof__ (uint64_t) != 0)
    {
      char *tmp = (char *) alloca (salt_len + __alignof__ (uint64_t));
      salt = copied_salt =
        memcpy (tmp + __alignof__ (uint64_t)
                - (tmp - (char *) 0) % __alignof__ (uint64_t),
                salt, salt_len);
      assert ((salt - (char *) 0) % __alignof__ (uint64_t) == 0);
    }


/* init bufs */
  bigbuf = alloca(6000);
  bigbuf2 = alloca(6000);
  bbl = bbl2 = 0;
  
  memcpy(bigbuf, key, key_len);
  bbl += key_len;
  memcpy(bigbuf+bbl, salt, salt_len);
  bbl += salt_len;
  
  memcpy(bigbuf2, key, key_len);
  bbl2 += key_len;
  memcpy(bigbuf2+bbl2, salt, salt_len);
  bbl2 += salt_len;
  memcpy(bigbuf2+bbl2, key, key_len);
  bbl2 += key_len;
  hash_sha512(bigbuf2, (char *)alt_result, bbl2);
  bbl2 = 0;

  for (cnt = key_len; cnt > 64; cnt -= 64)
  {
    memcpy(bigbuf+bbl, alt_result, 64);
    bbl += 64;
  }
  memcpy(bigbuf+bbl, alt_result, cnt);
  bbl += cnt;

  for (cnt = key_len; cnt > 0; cnt >>= 1)
    if ((cnt & 1) != 0)
    {
      memcpy(bigbuf+bbl, alt_result, 64);
      bbl += 64;
    }
    else
    {
      memcpy(bigbuf+bbl, key, key_len);
      bbl += key_len;
    }
  hash_sha512(bigbuf, (char *)alt_result, bbl);
  bbl = 0;


  for (cnt = 0; cnt < key_len; ++cnt)
  {
    memcpy(bigbuf2+bbl2, key, key_len);
    bbl2 += key_len;
  }
  hash_sha512(bigbuf2, (char *)temp_result, bbl2);
  bbl2 = 0;



  /* Create byte sequence P.  */
  cp = p_bytes = alloca (key_len);
  for (cnt = key_len; cnt >= 64; cnt -= 64)
    cp = memcpy (cp, temp_result, 64);
  memcpy (cp, temp_result, cnt);


  bbl2 = 0;
  for (cnt = 0; cnt < 16 + alt_result[0]; ++cnt)
  {
    memcpy((char *)(bigbuf2+bbl2), salt, salt_len);
    bbl2 += salt_len;
  }
  hash_sha512(bigbuf2, (char *)&temp_result, bbl2);
  bbl2=0;



  /* Create byte sequence S.  */
  cp = s_bytes = alloca (salt_len);
  for (cnt = salt_len; cnt >= 64; cnt -= 64)
    cp = memcpy (cp, temp_result, 64);
  memcpy (cp, temp_result, cnt);


  bbl=0;
  for (cnt = 0; cnt < rounds; ++cnt)
  {
      if ((cnt & 1) != 0)
      {
	memcpy(bigbuf+bbl, p_bytes, key_len);
	bbl += key_len;
      }
      else
      {
	memcpy(bigbuf+bbl, alt_result, 64);
	bbl += 64;
      }

      if (cnt % 3 != 0)
      {
	memcpy(bigbuf+bbl, s_bytes, salt_len);
	bbl += salt_len;
      }

      if (cnt % 7 != 0)
      {
	memcpy(bigbuf+bbl, p_bytes, key_len);
	bbl += key_len;
      }
      if ((cnt & 1) != 0)
      {
	memcpy(bigbuf+bbl, alt_result, 64);
	bbl += 64;
      }
      else
      {
	memcpy(bigbuf+bbl, p_bytes, key_len);
	bbl += key_len;
      }
      hash_sha512(bigbuf, (char *)alt_result, bbl);
      bbl = 0;
  }


  /* Now we can construct the result string.  It consists of three
     parts.  */
  cp = __stpncpy (buffer, sha512_salt_prefix, MAX (0, buflen));
  buflen -= sizeof (sha512_salt_prefix) - 1;

  if (rounds_custom)
    {
      int n = snprintf (cp, MAX (0, buflen), "%s%zu$",
                        sha512_rounds_prefix, rounds);
      cp += n;
      buflen -= n;
    }

  cp = __stpncpy (cp, salt, MIN ((size_t) MAX (0, buflen), salt_len));
  buflen -= MIN ((size_t) MAX (0, buflen), salt_len);

  if (buflen > 0)
    {
      *cp++ = '$';
      --buflen;
    }

  void b64_from_24bit (unsigned int b2, unsigned int b1, unsigned int b0,
                       int n)
  {
    unsigned int w = (b2 << 16) | (b1 << 8) | b0;
    while (n-- > 0 && buflen > 0)
      {
        *cp++ = b64t[w & 0x3f];
        --buflen;
        w >>= 6;
      }
  }


  b64_from_24bit (alt_result[0], alt_result[21], alt_result[42], 4);
  b64_from_24bit (alt_result[22], alt_result[43], alt_result[1], 4);
  b64_from_24bit (alt_result[44], alt_result[2], alt_result[23], 4);
  b64_from_24bit (alt_result[3], alt_result[24], alt_result[45], 4);
  b64_from_24bit (alt_result[25], alt_result[46], alt_result[4], 4);
  b64_from_24bit (alt_result[47], alt_result[5], alt_result[26], 4);
  b64_from_24bit (alt_result[6], alt_result[27], alt_result[48], 4);
  b64_from_24bit (alt_result[28], alt_result[49], alt_result[7], 4);
  b64_from_24bit (alt_result[50], alt_result[8], alt_result[29], 4);
  b64_from_24bit (alt_result[9], alt_result[30], alt_result[51], 4);
  b64_from_24bit (alt_result[31], alt_result[52], alt_result[10], 4);
  b64_from_24bit (alt_result[53], alt_result[11], alt_result[32], 4);
  b64_from_24bit (alt_result[12], alt_result[33], alt_result[54], 4);
  b64_from_24bit (alt_result[34], alt_result[55], alt_result[13], 4);
  b64_from_24bit (alt_result[56], alt_result[14], alt_result[35], 4);
  b64_from_24bit (alt_result[15], alt_result[36], alt_result[57], 4);
  b64_from_24bit (alt_result[37], alt_result[58], alt_result[16], 4);
  b64_from_24bit (alt_result[59], alt_result[17], alt_result[38], 4);
  b64_from_24bit (alt_result[18], alt_result[39], alt_result[60], 4);
  b64_from_24bit (alt_result[40], alt_result[61], alt_result[19], 4);
  b64_from_24bit (alt_result[62], alt_result[20], alt_result[41], 4);
  b64_from_24bit (0, 0, alt_result[63], 2);

  if (buflen <= 0)
    {
      __set_errno (ERANGE);
      buffer = NULL;
    }
  else
    *cp = '\0';         /* Terminate the string.  */

  return buffer;
}



