/* 
 * threads.c
 *
 * hashkill - a hash cracking tool
 * Copyright (C) 2010 Milen Rangelov <gat3way@gat3way.eu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


//#define DEBUG

#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/sysinfo.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <stdint.h>
#include <math.h>
#include <alloca.h>
#include <termios.h>
#include <openssl/bn.h>
#include <openssl/crypto.h>
#include "err.h"
#include "threads.h"
#include "hashinterface.h"
#include "plugins.h"
#include "sessions.h"


/* Global static variables */
static int attack_over;					// flag indicating attack is done
static int hash_len; 					// strlen(hash)
static uint64_t attack_overall_count;			// max. (attack metrics)
static uint64_t attack_current_count;			// cur. (attack metrics)
static uint64_t attack_current_length;			// bruteforce, current string len
static int *worker_thread_keys[HASHKILL_MAXTHREADS];	// used to set thread args
static char *attack_current_str;			// current bruteforce string
static uint64_t bruteforce_current_position;		// set by sessions.c
static int bruteforce_current_length;			// set by sessions.c
static long dictionary_current_pos;			// set by sessions.c
static int qflag;

/* OpenSSL thread-safety variables */
static pthread_mutex_t *lock_cs;			// mutexes
static long            *lock_count;			// mutex_count

static int pushready[HASHKILL_MAXTHREADS];
static int popready[HASHKILL_MAXTHREADS];
static int bruteforce_prefixlen;


/* Function prototypes */
static void * start_monitor_thread(void *arg);
static void SSL_thread_cleanup( void );
static void SSL_pthreads_locking_callback(int mode, int type, char *file, int line);
static unsigned long SSL_pthreads_thread_id(void);
static void SSL_thread_setup(void);
static inline void pushqueue(char *ptext, int queueid);
static inline char * popqueue(int queueid);
static void thread_wait_empty_queues(void);
static void thread_attack_worker_nonlinear(int *who);
static void thread_attack_worker_linear(int *who);
static void * worker_thread_action(void *arg);
hash_stat spawn_threads(unsigned int num, unsigned int queue_size);
unsigned int hash_num_cpu(void);
unsigned long hash_get_total_memory(void);
static void bruteforce_fillqueue(char *plaintext);
static void bruteforce_permute(char *str, int n, const int maxdepth, const char *baseset, const int numbase);
hash_stat main_thread_hybrid(void);
hash_stat main_thread_dictionary(void);
hash_stat main_thread_bruteforce(void);
void thread_attack_cleanup(void);
void thread_set_bf_len(int len);
void thread_set_bf_str(char *str);
void thread_set_dict_pos(long pos);
void thread_set_bf_pos(uint64_t pos);





/* Cleanup openssl thread-safe callbacks */
static void SSL_thread_cleanup(void) 
{
    int x;
    CRYPTO_set_locking_callback(NULL);
    for( x = 0; x < CRYPTO_num_locks(); x++ ) 
    {
	pthread_mutex_destroy(&(lock_cs[x]));
    }
    OPENSSL_free(lock_cs);
    OPENSSL_free(lock_count);
}


/* The openssl crypto locking callback */
static void SSL_pthreads_locking_callback(int mode, int type, char *file, int line) 
{
    if(mode & CRYPTO_LOCK)
    {
	pthread_mutex_lock(&(lock_cs[type]));
	lock_count[type]++;
    } 
    else
    { 
	pthread_mutex_unlock(&(lock_cs[type]));
    }
}


/* The next callback needed */
static unsigned long SSL_pthreads_thread_id(void) 
{
  unsigned long ret;
  ret = (unsigned long)pthread_self();

  return(ret);
}



/* setup open crypto callbacks */
static void SSL_thread_setup(void) 
{
    int x;

#define OPENSSL_THREAD_DEFINES
#include <openssl/opensslconf.h>
#if defined(THREADS) || defined(OPENSSL_THREADS)
#else
    fprintf(stderr, "WARNING: your openssl libraries were compiled without thread support\n");
    pthread_sleep_np( 2 );
#endif
    lock_cs    = (pthread_mutex_t*)OPENSSL_malloc(CRYPTO_num_locks()*sizeof(pthread_mutex_t));
    lock_count = (long*)OPENSSL_malloc(CRYPTO_num_locks() * sizeof(long));
    for( x = 0; x < CRYPTO_num_locks(); x++ )
    {
	lock_count[x] = 0;
	pthread_mutex_init(&(lock_cs[x]), NULL);
    }
    CRYPTO_set_id_callback((unsigned long (*)())SSL_pthreads_thread_id);
    CRYPTO_set_locking_callback((void (*)())SSL_pthreads_locking_callback);
}




/* Push plaintext into queue */
static inline void pushqueue(char *ptext, int queueid)
{
    int val;
    
    val = hash_thread_queue_counts[queueid];
    if (val == 0)
    {
	pthread_mutex_lock(&queuelock[queueid]);
	pushready[queueid] = 1;

    }

    memcpy(hash_queue[queueid][val].plaintext, ptext, 16);
    *(hash_queue[queueid][val].plaintext+16)=0;
    val++;
    hash_thread_queue_counts[queueid]++;

    if (val == hash_queue_size-1)
    {
	pushready[queueid] = 0;
	pthread_mutex_unlock(&queuelock[queueid]);
    }
}


/* 
    Pop plaintext from queue (flag indicates locking)
    queue mutex MUST be locked prior to calling if flag=1
*/
static inline char * popqueue(int queueid)
{
    int val=0;
    char *str;

    again:
    val = hash_thread_queue_counts[queueid];
    if (val==0) 
    {
	if (popready[queueid]==1)
	{
	    popready[queueid]=0;
	    pthread_mutex_unlock(&queuelock[queueid]);
	}
	else 
	{
	    usleep(1000);
	    goto again;
	}
    }
    else
    {
	if (popready[queueid]==0)
	{
	    pthread_mutex_lock(&queuelock[queueid]);
	    popready[queueid]=1;
	}
	str = hash_queue[queueid][val].plaintext;
	hash_thread_queue_counts[queueid]--;

	return str;
    }
    return hash_queue[queueid][0].plaintext;
}




/* Wait until queues have been emptied */
static void thread_wait_empty_queues(void)
{
    int cnt,a;
    unsigned int flag = 1;

    while (flag==1)
    {
	flag = 0;
	for (cnt=0;cnt<hash_num_threads;cnt++) 
	{
	    usleep(1000);
	    a = hash_thread_queue_counts[cnt];
	    if (a>0) flag = 1;
	}
    }
    attack_over = 2;
    for (cnt=0;cnt<hash_num_threads;cnt++) pthread_cancel(hash_threads[cnt]);
    SSL_thread_cleanup();
}



/* worker thread - nonlinear complexity */
static void thread_attack_worker_nonlinear(int *who)
{
    struct  hash_list_s  *mylist;
    struct hash_list_s * mylist1;
    int whoami;
    int val=0;
    char *plain = alloca(64);

    whoami = (int )*who;
    while (1)
    {
	if (attack_over == 2) pthread_exit(NULL);

	strcpy(plain, popqueue(whoami));
	hash_thread_processed[whoami]++;
	mylist = hash_list;
	while (mylist)
	{
		if ( (hash_plugin_check_hash(mylist->hash,plain,
		    mylist->salt,mylist->salt2, mylist->username) == hash_ok))
		{
			add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
			#ifdef DEBUG
			hlog("Cracked hash: %s plaintext: %s thread: %d\n",mylist->hash, plain, whoami);
			#endif
			if (mylist) del_hash_list(mylist);
		}
		pthread_mutex_lock(&listmutex);
	    	if (mylist) mylist = mylist->next;
		pthread_mutex_unlock(&listmutex);
	}
    }
    pthread_exit(NULL);
}




/* worker thread - linear complexity */
static void thread_attack_worker_linear(int *who)
{
    struct  hash_list_s  *mylist, *addlist;
    int whoami;
    unsigned int val=0;
    char myhash[HASHFILE_MAX_LINE_LENGTH];
    char *plain = alloca(64);
    int flag = 0;

    whoami = (int )*who;
    while (1)
    {

	strcpy(plain, popqueue(whoami));

	mylist = hash_list;
	if (mylist)
	{
	    (void)hash_plugin_check_hash(mylist->hash, plain,
		    mylist->salt, myhash, mylist->username);
	}
	hash_thread_processed[whoami]++;
	while (mylist)
	{
		if (memcmp(mylist->hash, myhash, strlen(myhash))==0)
		{
			#ifdef DEBUG
			hlog("Cracked hash: %s plaintext: %s thread: %d\n",mylist->hash, plain, whoami);
			#endif
			flag = 0;
			pthread_mutex_lock(&crackedmutex);
			addlist = cracked_list;
			while (addlist)
			{
			    if ( (strcmp(addlist->username, mylist->username) == 0) && (strcmp(addlist->hash, mylist->hash) == 0)
				&& (strcmp(addlist->salt, mylist->salt) == 0))
				    flag = 1;
			    addlist = addlist->next;
			}
			pthread_mutex_unlock(&crackedmutex);
			if (flag == 0) add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
		}
	    	if (mylist) mylist = mylist->next;
	}
    }
}




/* Worker thread entrypoint */
static void * worker_thread_action(void *arg)
{
    int flag = 0;
    unsigned int complex;
    struct  hash_list_s *amylist;
    cpu_set_t set;

    memcpy(&complex,arg,sizeof(int));
    free(worker_thread_keys[complex]);
    amylist = hash_list;
    while (amylist)
    {
	if (strlen(amylist->salt)>1) flag = 1;
	amylist = amylist->next;
    }

    if (flag == 0)
    {
	if (complex == 1) hlog("Attack has linear complexity%s\n","");
	thread_attack_worker_linear(&complex);
    }
    else
    {
	if (complex == 1) hlog("Attack has non-linear complexity%s\n","");
	thread_attack_worker_nonlinear(&complex);
    }

    pthread_exit(NULL);
    return 0;
}



/* Spawn worker threads, monitor thread */
hash_stat spawn_threads(unsigned int num, unsigned int queue_size)
{
    unsigned int cnt, cnt1;
    int status, status1 = 0;
    pthread_t monitorthread;
    attack_over = 0;

    qflag = 0;
    hash_len = hash_plugin_hash_length();
    if (num < 1) return hash_err;
    if (queue_size > HASHKILL_MAXQUEUESIZE) return hash_err;
    if (init_mutexes() == hash_err) return hash_err;
    
    for (cnt=0; cnt<num; cnt++)
    {
	for (cnt1=0; cnt1 < queue_size; cnt1++)
	{
	    hash_queue[cnt][cnt1].plaintext = malloc(HASHFILE_MAX_PLAIN_LENGTH+1);
	    bzero(hash_queue[cnt][cnt1].plaintext,HASHFILE_MAX_PLAIN_LENGTH+1); // make valgrind happy
	    pushready[cnt] = 0;
	    popready[cnt] = 0;

	    if (!hash_queue[cnt][cnt1].plaintext)
	    {
		elog("Not enough memory to create thread queues (thread: %d fifo_elem: %d)", cnt, cnt1);
		status1 = 1;
		break;
	    }
	}
	SSL_thread_setup();
	
	hash_thread_queue_counts[cnt]=0;
	hash_thread_processed[cnt]=0;
	hash_thread_cracked[cnt]=0;
	worker_thread_keys[cnt] = malloc(sizeof(int));
	*(worker_thread_keys[cnt]) = cnt;
	status = pthread_create(&hash_threads[cnt], NULL, worker_thread_action, (void *)worker_thread_keys[cnt]);

	if (status != 0) elog("Cannot spawn thread %d\n",cnt);
	if ((status != 0) || (status1 == 1))
	{
	    for (cnt=0; cnt < num; cnt++)
		for (cnt1=0; cnt1 < queue_size; cnt1++)
		    if (hash_queue[cnt][cnt1].plaintext != NULL) free(hash_queue[cnt][cnt1].plaintext);
	    return hash_err;
	}
    }
    hash_num_threads = num;
    hash_queue_size = queue_size;
    hlog("Spawned %d threads\n", hash_num_threads);
    cnt++;
    pthread_create(&monitorthread, NULL, start_monitor_thread, &cnt);
    hlog("Spawned monitor thread%s\n", "");
    return hash_ok;
}



/* Get number of CPUs, read from /proc/cpuinfo */
unsigned int hash_num_cpu(void)
{
    FILE *fcpu;
    unsigned int proc=0;
    char buf[128];
    char *status;

    fcpu=fopen("/proc/cpuinfo","r");
    if (!fcpu)
    {
        elog("Cannot open /proc/cpuinfo! errno=%d", errno);
        return 0;
    }
    else
    {
        do
        {
            status = fgets(buf, 128, fcpu);
            if ((status) && (strstr(buf, "processor"))) proc++;
        } while (status);

        fclose(fcpu);
    }
    hlog("Detected %d CPUs.\n", proc);
    return proc;
}



/* Get physical memory in MB */
unsigned long hash_get_total_memory(void)
{
    struct sysinfo info;
    unsigned long memory_mb;

    sysinfo(&info);
    memory_mb = ((info.totalram / info.mem_unit) / (1024*1024));
    hlog("Total physical memory : %ld MB\n", memory_mb);
    return memory_mb;
}



/* Initialize synchronization primitives */
hash_stat init_mutexes(void)
{
    int cnt;
    pthread_mutexattr_t mutexattr;

    pthread_mutexattr_init(&mutexattr); 
    pthread_mutexattr_settype(&mutexattr, PTHREAD_MUTEX_NORMAL); 
    pthread_mutexattr_setpshared(&mutexattr, PTHREAD_PROCESS_PRIVATE);
    pthread_mutexattr_setprotocol(&mutexattr, PTHREAD_PRIO_NONE);

    for (cnt=0;cnt<=hash_num_threads;cnt++)
    {
	pthread_mutexattr_init(&mutexattr); 
	pthread_mutexattr_settype(&mutexattr, PTHREAD_MUTEX_NORMAL); 
	pthread_mutexattr_setpshared(&mutexattr, PTHREAD_PROCESS_SHARED);
	pthread_mutexattr_setprotocol(&mutexattr, PTHREAD_PRIO_NONE);
	
	if (pthread_mutex_init(&queuelock[cnt], &mutexattr)!=0)
	{
    	    elog("Cannot create queue mutex %s\n","");
    	    return hash_err;
	}
    }
    if (pthread_mutex_init(&listmutex, &mutexattr))
    {
        elog("Cannot create list mutex %s\n","");
        return hash_err;
    }
    if (pthread_mutex_init(&crackedmutex, &mutexattr))
    {
        elog("Cannot create list mutex %s\n", "");
        return hash_err;
    }
    hlog("Initialized thread mutexes%s\n", "");
    return hash_ok;
}



/* Start monitor thread */ 
static void * start_monitor_thread(void *arg)
{

    usleep(1000); 
    printf("\nAttack statistics...\n");
#ifdef DEBUG
    int overall, o1,o2,o3,o4;
    int p1,p2,p3,p4;

    sleep(2);
    while (attack_over == 0)
    {
	overall = 0;
	o1=hash_thread_processed[0]/2;
	o2=hash_thread_processed[1]/2;
	o3=hash_thread_processed[2]/2;
	o4=hash_thread_processed[3]/2;
	hash_thread_processed[0]=0;
	hash_thread_processed[1]=0;
	hash_thread_processed[2]=0;
	hash_thread_processed[3]=0;
	p1 = hash_thread_queue_counts[0];
	p2 = hash_thread_queue_counts[1];
	p3 = hash_thread_queue_counts[2];
	p4 = hash_thread_queue_counts[3];




	printf("(procsd): %d  %d  %d  %d\n",o1,o2,o3,o4);
	printf("(queues): %d  %d  %d  %d\n",p1,p2,p3,p4);
	sleep(2);

    }
    printf("\n");
    return 0;
#else
    int sum,cracked;
    int cnt;
    struct hash_list_s *mylist;
    FILE *sessionfile;
    char key[2];

    
    qflag = 0;
    while (attack_over == 0)
    {
	sleep(2);
	qflag++;
	if (qflag>4) qflag = 4;
	sum = 0;
	if (attack_over == 0)
	{
	    /* show stats */
	    for (cnt = 0; cnt < hash_num_threads; cnt++)
	    {
		sum += hash_thread_processed[cnt];
		hash_thread_processed[cnt]=0;
	    }
	    cracked = 0;
	    pthread_mutex_lock(&crackedmutex);
	    mylist = cracked_list;
	    while (mylist)
	    {
		mylist = mylist->next;
		cracked++;
	    }
	    
	    pthread_mutex_unlock(&crackedmutex);
	    /* This is stupid I know, should be fixed some day */
	    if ((sum / 2000)>5)
	    {
		if ( ((attack_current_count*100)/attack_overall_count) > 100)
		{
		    printf("\rProgress: 100%%   Speed: %d KPlaintexts/sec   Cracked: %d hashes   ", (sum / 2000), cracked);
		}
		else  printf("\rProgress: %lld%%   Speed: %d KPlaintexts/sec   Cracked: %d hashes   ", ((attack_current_count*100)/attack_overall_count) ,(sum / 2000), cracked);
	    }
	    else
	    {
		if ( ((attack_current_count*100)/attack_overall_count) > 100)
		{
		    printf("\rProgress: 100%%   Speed: %d plaintexts/sec   Cracked: %d hashes   ", (sum / 2), cracked);
		}
		else  printf("\rProgress: %lld%%   Speed: %d plaintexts/sec   Cracked: %d hashes   ", ((attack_current_count*100)/attack_overall_count) ,(sum / 2), cracked);
	    }
	    fflush(stdout);

	    char key[32];

	    fcntl(0, F_SETFL,O_NONBLOCK);
	    read(0, &key,2);
	    fcntl(0, F_SETFL,~O_NONBLOCK);
	    if ((key[0]=='\n')||(key[1]=='\n'))
	    {
		printf("\n");
		print_cracked_list();
		fflush(stdout);
	    }
	    key[0]=0;key[1]=0;
	    
	    /* save session : TBD: check error status*/
	    if (session_init_file(&sessionfile) == hash_ok)
	    {
		session_write_parameters(get_current_plugin(), attack_method, ((attack_current_count*100)/attack_overall_count), sessionfile);
		if (attack_method == attack_method_dictionary)
		{
		    session_write_dictionary_parm(attack_current_count, hash_dictionary_file, sessionfile);
		}
		else if (attack_method == attack_method_simple_bruteforce)
		{
		    session_write_bruteforce_parm(bruteforce_start, bruteforce_end, bruteforce_prefix, bruteforce_suffix, bruteforce_charset, attack_current_length, attack_current_str, attack_current_count, sessionfile);
		}
		else if (attack_method == attack_method_hybrid)
		{
		    session_write_hybrid_parm(hash_dictionary_file, bruteforce_start, bruteforce_end, bruteforce_charset, attack_current_count, attack_current_str,  sessionfile);
		}

		session_write_hashlist(sessionfile);
		session_write_crackedlist(sessionfile);
		session_close_file(sessionfile);
	    }
	    if (cracked >= hashes_count) attack_over = 2;
	}

    }
    printf("\n");
    pthread_exit(NULL);
    return 0;
#endif
}



/* Main thread - bruteforce attack - feed workers */
static void bruteforce_fillqueue(char *plaintext)
{
    static unsigned int cnt=0;
    int flag;
    int val;
    int len;
    
    if ((plaintext)&&(attack_over==0))
    {
        len = strlen(plaintext);
        *(plaintext+len)=0;
        attack_current_str = plaintext;
        if (attack_method != attack_method_hybrid) attack_current_count++;
        do
        {
	    flag = 0;
	    cnt++;
	    if (cnt == hash_num_threads) cnt = 0;
	    if ((pushready[cnt]==1) || ((pushready[cnt]==0)&&(hash_thread_queue_counts[cnt]==0)))
	    {
		pushqueue(plaintext,cnt);
		flag = 1;
	    }
	} while (flag == 0);
    }
}



/* Main thread - bruteforce attack - recursive permutations generator */
static void bruteforce_permute(char *str, int n, const int maxdepth, const char *baseset, const int numbase)
{
    register int i;
    char *currpos = &str[n-1];
    const char *currchar = baseset;
    
    if (attack_over == 0)
    {
	if (n<maxdepth)
	{
    	    for (i=0; i<numbase; i++)
    	    {
        	*currpos = *currchar++;
        	bruteforce_permute(str, n+1, maxdepth, baseset, numbase);
    	    }
	}
	else
	{
    	    for (i=0; i<numbase; i++)
    	    {
        	*currpos = *currchar++;
        	if (bruteforce_suffix[0]!=0)
        	{
        	    strcat(str,bruteforce_suffix);
		    bruteforce_fillqueue(str-bruteforce_prefixlen);
		    *(str+maxdepth) = 0;
		}
		else bruteforce_fillqueue(str-bruteforce_prefixlen);
    	    }
	}
    }
}



/* Main thread - hybrid attack */
hash_stat main_thread_hybrid(void)
{
    FILE *dictfile;
    char *buf;
    int cnt;
    int flag,val,len;
    time_t time1,time2;
    struct stat filestat;

    attack_over = 0;
    time1 = time(NULL);
    hlog("Launching hybrid attack\n%s","");

    dictfile = fopen(hash_dictionary_file, "r");
    if (!dictfile)
    {
	elog("Cannot open dictionary file \'%s\'\n", hash_dictionary_file);
	return hash_err;
    }
    
    stat(hash_dictionary_file, &filestat);
    attack_overall_count = filestat.st_size;
    attack_current_count = 0;
    
    /* we are loading a session file? */
    if (session_restore_flag == 1)
    {
	if (fseek(dictfile, dictionary_current_pos, SEEK_SET)!=0)
	{
	    elog("Cannot seek to position %ld in dictionary file!\n",dictionary_current_pos);
	}
	attack_current_count = dictionary_current_pos;
    }
    
    buf = malloc(HASHFILE_MAX_PLAIN_LENGTH+1);
    cnt = 0;
    flag = 0;

    do
    {
	fgets_unlocked(buf,HASHFILE_MAX_LINE_LENGTH,dictfile);
	len = strlen(buf);
	val=0;
	if (buf)
	{

	    attack_current_count += len;
	    bzero(buf+len-1, 1);
	    strcpy(bruteforce_prefix, buf);
	    do
	    {
		flag = 0;
		cnt++;
		if (cnt == hash_num_threads) cnt = 0;

		if ((pushready[cnt]==1) || ((pushready[cnt]==0)&&(hash_thread_queue_counts[cnt]==0)))
		{
		    pushqueue(buf, cnt);
		    flag = 1;
		}

	    } while (flag == 0);

	    /* next try all combinations */
	    for (len = bruteforce_start; len <= bruteforce_end; len++)
	    {
		if (attack_over > 0) break;
		bruteforce_permute((buf+strlen(buf)), 1, len, bruteforce_charset, strlen(bruteforce_charset));
	    }
	    
	}
    } while ((!feof(dictfile))&&(hash_list)&&(attack_over == 0));

    for (cnt=0; cnt<hash_num_threads; cnt++) pthread_mutex_unlock(&queuelock[cnt]);
    free(buf);
    thread_wait_empty_queues();
    time2 = time(NULL);
    printf("\n");
    hlog("Hybrid attack complete. It took %ld seconds...\n",time2-time1);
    return hash_ok;

}

/* Main thread - bruteforce attack function */
hash_stat main_thread_bruteforce(void)
{
    char *buf;
    int len;
    time_t time1,time2;
    int cnt;

    attack_over = 0;
    time1 = time(NULL);
    bruteforce_prefixlen = strlen(bruteforce_prefix);
    hlog("Launching bruteforce attack\n%s","");
    /* we are not restoring a session? */
    if (session_restore_flag == 0)
    {
	attack_overall_count = 1;
	for (cnt = bruteforce_start; cnt <= bruteforce_end; cnt++)
	{
	    attack_current_length = cnt;
	    attack_overall_count += (powl(strlen(bruteforce_charset),cnt));
	}
	attack_current_count = 1;
	buf = malloc(HASHFILE_MAX_PLAIN_LENGTH+1);
	bzero(buf, HASHFILE_MAX_PLAIN_LENGTH+1); // make valgrind happy
	for (len = bruteforce_start; len <= bruteforce_end; len++)
	{
	    if (attack_over > 0) goto out;
	    strcpy(buf,bruteforce_prefix);
	    bruteforce_permute((buf+strlen(bruteforce_prefix)), 1, len, bruteforce_charset, strlen(bruteforce_charset));
	}
    }
    /* we are restoring a session then */
    else
    {
	attack_overall_count = 1;
	for (cnt = bruteforce_start; cnt <= bruteforce_end; cnt++)
	{
	    attack_current_length = cnt;
	    attack_overall_count += powl(strlen(bruteforce_charset),cnt);
	}
	attack_current_count = bruteforce_current_position;
	buf = malloc(HASHFILE_MAX_PLAIN_LENGTH+1);
	bzero(buf, HASHFILE_MAX_PLAIN_LENGTH+1); // make valgrind happy
	strcpy(buf, bruteforce_prefix);
	strcat(buf, attack_current_str);
	int temp=1;
	while (attack_current_str[temp] == bruteforce_charset[strlen(bruteforce_charset)]) temp++;
	
	bruteforce_permute((buf+strlen(bruteforce_prefix)), temp, attack_current_length, bruteforce_charset, strlen(bruteforce_charset));
	for (len = attack_current_length+1; len <= bruteforce_end; len++)
	{
	    if (attack_over > 0) goto out;
	    strcpy(buf,bruteforce_prefix);
	    bruteforce_permute((buf+strlen(bruteforce_prefix)), 1, len, bruteforce_charset, strlen(bruteforce_charset));
	}
    }
    
    out:
    for (cnt=0; cnt<hash_num_threads; cnt++) pthread_mutex_unlock(&queuelock[cnt]);
    free(buf);
    thread_wait_empty_queues();
    time2 = time(NULL);
    printf("\n");
    hlog("Bruteforce attack complete. It took %ld seconds...\n",time2-time1);
    return hash_ok;
}




/* Main thread - dictionary attack function */
hash_stat main_thread_dictionary(void)
{
    FILE *dictfile;
    char *buf;
    int cnt;
    int flag,val,len;
    time_t time1,time2;
    struct stat filestat;

    attack_over = 0;
    time1 = time(NULL);
    hlog("Launching dictionary attack\n%s","");

    dictfile = fopen(hash_dictionary_file, "r");
    if (!dictfile)
    {
	elog("Cannot open dictionary file \'%s\'\n", hash_dictionary_file);
	return hash_err;
    }
    
    stat(hash_dictionary_file, &filestat);
    attack_overall_count = filestat.st_size;
    attack_current_count = 0;
    
    /* we are loading a session file? */
    if (session_restore_flag == 1)
    {
	if (fseek(dictfile, dictionary_current_pos, SEEK_SET)!=0)
	{
	    elog("Cannot seek to position %ld in dictionary file!\n",dictionary_current_pos);
	}
	attack_current_count = dictionary_current_pos;
    }

    buf = malloc(HASHFILE_MAX_PLAIN_LENGTH+1);
    cnt = 0;
    flag = 0;
    do
    {
	fgets_unlocked(buf, HASHFILE_MAX_PLAIN_LENGTH, dictfile);
	len = strlen(buf);
	val=0;
	attack_current_count += len;

	if (buf)
	{

	    bzero(buf+len-1, 1);
	    do
	    {
		flag = 0;
		cnt++;
		if (cnt == hash_num_threads) cnt = 0;
		if ((pushready[cnt]==1) || ((pushready[cnt]==0)&&(hash_thread_queue_counts[cnt]==0)))
		{
		    pushqueue(buf,cnt);
		    flag = 1;
		}

	    } while (flag == 0);
	}
    } while ((!feof(dictfile))&&(attack_over == 0));

    for (cnt=0; cnt<hash_num_threads; cnt++) pthread_mutex_unlock(&queuelock[cnt]);
    free(buf);
    thread_wait_empty_queues();
    time2 = time(NULL);
    attack_over = 2;

    printf("\n");
    hlog("Dictionary attack complete. It took %ld seconds...\n",time2-time1);
    return hash_ok;
}


/* free memory used by queues */
void thread_attack_cleanup(void)
{
    int cnt, cnt1;
    
    for (cnt = 0; cnt <= hash_num_threads; cnt++)
	for (cnt1 = 0; cnt1 < hash_queue_size; cnt1++)
	    if (hash_queue[cnt][cnt1].plaintext)
		free(hash_queue[cnt][cnt1].plaintext);
    attack_over = 2;
}



/* Used by sessions.c to set bf params */
void thread_set_bf_len(int len)
{
    bruteforce_current_length = len;
}



/* Used by sessions.c to set bf params */
void thread_set_bf_str(char *str)
{
    attack_current_str = malloc(strlen(str)+1);
    strcpy(attack_current_str, str);
}



/* Used by sessions.c to set dict params */
void thread_set_dict_pos(long pos)
{
    dictionary_current_pos = pos;
}



/* Used by sessions.c to set bf params */
void thread_set_bf_pos(uint64_t pos)
{
    bruteforce_current_position = pos;
}
